@@ -343,7 +343,7 @@ check_server_version(unsigned long version_number,
 }
 
 /*********************************************************************//**
-Receive options important for XtraBackup from MySQL server.
+Receive options important for XtraBackup from server.
 @return	true on success. */
 bool get_mysql_vars(MYSQL *connection)
 {
@@ -1843,8 +1843,8 @@ static std::string make_local_paths(const char *data_file_path)
 bool write_backup_config_file()
 {
 	int rc= backup_file_printf("backup-my.cnf",
-		"# This MySQL options file was generated by innobackupex.\n\n"
-		"# The MySQL server\n"
+		"# This options file was generated by innobackupex.\n\n"
+		"# The server\n"
 		"[mysqld]\n"
 		"innodb_checksum_algorithm=%s\n"
 		"innodb_data_file_path=%s\n"
@@ -1929,7 +1929,7 @@ flush_changed_page_bitmaps()
 
 
 /*********************************************************************//**
-Deallocate memory, disconnect from MySQL server, etc.
+Deallocate memory, disconnect from server, etc.
 @return	true on success. */
 void
 backup_cleanup()
