@@ -43,7 +43,7 @@ C_MODE_START
 extern unsigned int mysql_server_last_errno;
 extern char mysql_server_last_error[MYSQL_ERRMSG_SIZE];
 static my_bool emb_read_query_result(MYSQL *mysql);
-static void emb_free_embedded_thd(MYSQL *mysql);
+static void free_embedded_thd(MYSQL *mysql);
 static bool embedded_print_errors= 0;
 
 extern "C" void unireg_clear(int exit_code)
@@ -115,13 +115,13 @@ emb_advanced_command(MYSQL *mysql, enum enum_server_command command,
   NET *net= &mysql->net;
   my_bool stmt_skip= stmt ? stmt->state != MYSQL_STMT_INIT_DONE : FALSE;
 
-  if (thd->killed != NOT_KILLED)
+  if (thd && thd->killed != NOT_KILLED)
   {
     if (thd->killed < KILL_CONNECTION)
       thd->killed= NOT_KILLED;
     else
     {
-      emb_free_embedded_thd(mysql);
+      free_embedded_thd(mysql);
       thd= 0;
     }
   }
@@ -430,7 +430,7 @@ int emb_unbuffered_fetch(MYSQL *mysql, char **row)
   return 0;
 }
 
-static void emb_free_embedded_thd(MYSQL *mysql)
+static void free_embedded_thd(MYSQL *mysql)
 {
   THD *thd= (THD*)mysql->thd;
   server_threads.erase(thd);
@@ -453,12 +453,23 @@ static MYSQL_RES * emb_store_result(MYSQL *mysql)
   return mysql_store_result(mysql);
 }
 
-int emb_read_change_user_result(MYSQL *mysql)
+static int emb_read_change_user_result(MYSQL *mysql)
 {
   mysql->net.read_pos= (uchar*)""; // fake an OK packet
   return mysql_errno(mysql) ? (int)packet_error : 1 /* length of the OK packet */;
 }
 
+static void emb_on_close_free(MYSQL *mysql)
+{
+  my_free(mysql->info_buffer);
+  mysql->info_buffer= 0;
+  if (mysql->thd)
+  {
+    free_embedded_thd(mysql);
+    mysql->thd= 0;
+  }
+}
+
 MYSQL_METHODS embedded_methods= 
 {
   emb_read_query_result,
@@ -468,12 +479,12 @@ MYSQL_METHODS embedded_methods=
   emb_fetch_lengths, 
   emb_flush_use_result,
   emb_read_change_user_result,
+  emb_on_close_free,
   emb_list_fields,
   emb_read_prepare_result,
   emb_stmt_execute,
   emb_read_binary_rows,
   emb_unbuffered_fetch,
-  emb_free_embedded_thd,
   emb_read_statistics,
   emb_read_query_result,
   emb_read_rows_from_cursor
@@ -485,7 +496,7 @@ MYSQL_METHODS embedded_methods=
 
 char **copy_arguments(int argc, char **argv)
 {
-  uint length= 0;
+  size_t length= 0;
   char **from, **res, **end= argv+argc;
 
   for (from=argv ; from != end ; from++)
@@ -578,7 +589,7 @@ int init_embedded_server(int argc, char **argv, char **groups)
 
   /* Get default temporary directory */
   opt_mysql_tmpdir=getenv("TMPDIR");	/* Use this if possible */
-#if defined(__WIN__)
+#if defined(_WIN32)
   if (!opt_mysql_tmpdir)
     opt_mysql_tmpdir=getenv("TEMP");
   if (!opt_mysql_tmpdir)
@@ -663,7 +674,7 @@ void end_embedded_server()
 }
 
 
-void init_embedded_mysql(MYSQL *mysql, int client_flag)
+void init_embedded_mysql(MYSQL *mysql, ulong client_flag)
 {
   THD *thd = (THD *)mysql->thd;
   thd->mysql= mysql;
@@ -683,7 +694,7 @@ void init_embedded_mysql(MYSQL *mysql, int client_flag)
   create_new_thread(), and prepare_new_connection_state().  This should
   be refactored to avoid code duplication.
 */
-void *create_embedded_thd(int client_flag)
+void *create_embedded_thd(ulong client_flag)
 {
   THD * thd= new THD(next_thread_id());
 
@@ -699,7 +710,7 @@ void *create_embedded_thd(int client_flag)
   thd->set_command(COM_SLEEP);
   thd->set_time();
   thd->init_for_queries();
-  thd->client_capabilities= client_flag;
+  thd->client_capabilities= client_flag | MARIADB_CLIENT_EXTENDED_METADATA;
   thd->real_id= pthread_self();
 
   thd->db= null_clex_str;
@@ -830,7 +841,7 @@ int check_embedded_connection(MYSQL *mysql, const char *db)
   /* acl_authenticate() takes the data from thd->net->read_pos */
   thd->net.read_pos= (uchar*)buf;
 
-  if (acl_authenticate(thd, 0, end - buf))
+  if (acl_authenticate(thd, (uint) (end - buf)))
   {
     my_free(thd->security_ctx->user);
     goto err;
@@ -1100,11 +1111,11 @@ bool Protocol_text::store_field_metadata(const THD * thd,
   client_field->flags= (uint16) server_field.flags;
   client_field->decimals= server_field.decimals;
 
-  client_field->db_length=		strlen(client_field->db);
-  client_field->table_length=		strlen(client_field->table);
-  client_field->name_length=		strlen(client_field->name);
-  client_field->org_name_length=	strlen(client_field->org_name);
-  client_field->org_table_length=	strlen(client_field->org_table);
+  client_field->db_length=		(uint)strlen(client_field->db);
+  client_field->table_length=		(uint)strlen(client_field->table);
+  client_field->name_length=		(uint)strlen(client_field->name);
+  client_field->org_name_length=	(uint)strlen(client_field->org_name);
+  client_field->org_table_length=	(uint)strlen(client_field->org_table);
 
   client_field->catalog= dup_str_aux(field_alloc, "def", 3, cs, thd_cs);
   client_field->catalog_length= 3;
@@ -1378,12 +1389,12 @@ bool Protocol::net_store_data(const uchar *from, size_t length)
 
   if (!(field_buf= (char*) alloc_root(alloc, length + sizeof(uint) + 1)))
     return TRUE;
-  *(uint *)field_buf= length;
+  *(uint *)field_buf= (uint)length;
   *next_field= field_buf + sizeof(uint);
   memcpy((uchar*) *next_field, from, length);
   (*next_field)[length]= 0;
   if (next_mysql_field->max_length < length)
-    next_mysql_field->max_length=length;
+    next_mysql_field->max_length=(ulong)length;
   ++next_field;
   ++next_mysql_field;
   return FALSE;
@@ -1393,7 +1404,7 @@ bool Protocol::net_store_data(const uchar *from, size_t length)
 bool Protocol::net_store_data_cs(const uchar *from, size_t length,
                               CHARSET_INFO *from_cs, CHARSET_INFO *to_cs)
 {
-  uint conv_length= to_cs->mbmaxlen * length / from_cs->mbminlen;
+  size_t conv_length= length * to_cs->mbmaxlen / from_cs->mbminlen;
   uint dummy_error;
   char *field_buf;
   if (!thd->mysql)            // bootstrap file handling
@@ -1404,10 +1415,10 @@ bool Protocol::net_store_data_cs(const uchar *from, size_t length,
   *next_field= field_buf + sizeof(uint);
   length= copy_and_convert(*next_field, conv_length, to_cs,
                            (const char*) from, length, from_cs, &dummy_error);
-  *(uint *) field_buf= length;
+  *(uint *) field_buf= (uint)length;
   (*next_field)[length]= 0;
   if (next_mysql_field->max_length < length)
-    next_mysql_field->max_length= length;
+    next_mysql_field->max_length= (ulong)length;
   ++next_field;
   ++next_mysql_field;
   return false;
