@@ -472,7 +472,7 @@ priority of the background thread so that it will be scheduled and it
 can release the resource.  This solution is called priority inheritance
 in real-time programming.  A drawback of this solution is that the overhead
 of acquiring a mutex increases slightly, maybe 0.2 microseconds on a 100
-MHz Pentium, because the thread has to call os_thread_get_curr_id.  This may
+MHz Pentium, because the thread has to call pthread_self.  This may
 be compared to 0.5 microsecond overhead for a mutex lock-unlock pair. Note
 that the thread cannot store the information in the resource , say mutex,
 itself, because competing threads could wipe out the information if it is
@@ -514,7 +514,7 @@ static srv_sys_t	srv_sys;
 struct purge_coordinator_state
 {
   /** Snapshot of the last history length before the purge call.*/
-  uint32 m_history_length;
+  size_t m_history_length;
   Atomic_counter<int> m_running;
 private:
   ulint count;
@@ -655,7 +655,6 @@ static void srv_init()
 
 	/* Initialize some INFORMATION SCHEMA internal structures */
 	trx_i_s_cache_init(trx_i_s_cache);
-
 }
 
 /*********************************************************************//**
@@ -711,8 +710,6 @@ static void srv_refresh_innodb_monitor_stats(time_t current_time)
 	btr_cur_n_non_sea_old = btr_cur_n_non_sea;
 #endif /* BTR_CUR_HASH_ADAPT */
 
-	log_refresh_stats();
-
 	buf_refresh_io_stats();
 
 	srv_n_rows_inserted_old = srv_stats.n_rows_inserted;
@@ -1024,9 +1021,6 @@ srv_export_innodb_status(void)
 	export_vars.innodb_buffer_pool_read_requests
 		= buf_pool.stat.n_page_gets;
 
-	export_vars.innodb_buffer_pool_write_requests =
-		srv_stats.buf_pool_write_requests;
-
 	export_vars.innodb_buffer_pool_reads = srv_stats.buf_pool_reads;
 
 	export_vars.innodb_buffer_pool_read_ahead_rnd =
@@ -1074,7 +1068,7 @@ srv_export_innodb_status(void)
 		- UT_LIST_GET_LEN(buf_pool.free);
 
 	export_vars.innodb_max_trx_id = trx_sys.get_max_trx_id();
-	export_vars.innodb_history_list_length = trx_sys.history_size();
+	export_vars.innodb_history_list_length = trx_sys.history_size_approx();
 
 	mysql_mutex_lock(&lock_sys.wait_mutex);
 	export_vars.innodb_row_lock_waits = lock_sys.get_wait_cumulative();
@@ -1115,8 +1109,6 @@ srv_export_innodb_status(void)
 		srv_truncated_status_writes;
 
 	export_vars.innodb_page_compression_saved = srv_stats.page_compression_saved;
-	export_vars.innodb_index_pages_written = srv_stats.index_pages_written;
-	export_vars.innodb_non_index_pages_written = srv_stats.non_index_pages_written;
 	export_vars.innodb_pages_page_compressed = srv_stats.pages_page_compressed;
 	export_vars.innodb_page_compressed_trim_op = srv_stats.page_compressed_trim_op;
 	export_vars.innodb_pages_page_decompressed = srv_stats.pages_page_decompressed;
@@ -1161,19 +1153,17 @@ srv_export_innodb_status(void)
 			crypt_stat.estimated_iops;
 		export_vars.innodb_encryption_key_requests =
 			srv_stats.n_key_requests;
-		export_vars.innodb_key_rotation_list_length =
-			srv_stats.key_rotation_list_length;
 	}
 
 	mysql_mutex_unlock(&srv_innodb_monitor_mutex);
 
-	mysql_mutex_lock(&log_sys.mutex);
+	log_sys.latch.rd_lock(SRW_LOCK_CALL);
 	export_vars.innodb_lsn_current = log_sys.get_lsn();
 	export_vars.innodb_lsn_flushed = log_sys.get_flushed_lsn();
 	export_vars.innodb_lsn_last_checkpoint = log_sys.last_checkpoint_lsn;
 	export_vars.innodb_checkpoint_max_age = static_cast<ulint>(
 		log_sys.max_checkpoint_age);
-	mysql_mutex_unlock(&log_sys.mutex);
+	log_sys.latch.rd_unlock();
 	export_vars.innodb_os_log_written = export_vars.innodb_lsn_current
 		- recv_sys.lsn;
 
@@ -1596,7 +1586,7 @@ static bool srv_purge_should_exit()
     return true;
 
   /* Slow shutdown was requested. */
-  const uint32_t history_size= trx_sys.history_size();
+  const size_t history_size= trx_sys.history_size();
   if (history_size)
   {
     static time_t progress_time;
@@ -1606,9 +1596,9 @@ static bool srv_purge_should_exit()
       progress_time= now;
 #if defined HAVE_SYSTEMD && !defined EMBEDDED_LIBRARY
       service_manager_extend_timeout(INNODB_EXTEND_TIMEOUT_INTERVAL,
-				     "InnoDB: to purge %u transactions",
+				     "InnoDB: to purge %zu transactions",
 				     history_size);
-      ib::info() << "to purge " << history_size << " transactions";
+      sql_print_information("InnoDB: to purge %zu transactions", history_size);
 #endif
     }
     return false;
@@ -1818,10 +1808,10 @@ void purge_coordinator_state::refresh(bool full)
     lsn_hwm= adaptive_purge_threshold + series[n_threads];
   }
 
-  mysql_mutex_lock(&log_sys.mutex);
+  log_sys.latch.rd_lock(SRW_LOCK_CALL);
   const lsn_t last= log_sys.last_checkpoint_lsn,
     max_age= log_sys.max_checkpoint_age;
-  mysql_mutex_unlock(&log_sys.mutex);
+  log_sys.latch.rd_unlock();
 
   lsn_age_factor= ulint(((log_sys.get_lsn() - last) * 100) / max_age);
 }
@@ -1905,7 +1895,7 @@ static void srv_shutdown_purge_tasks()
   std::unique_lock<std::mutex> lk(purge_thd_mutex);
   while (!purge_thds.empty())
   {
-    innobase_destroy_background_thd(purge_thds.front());
+    destroy_background_thd(purge_thds.front());
     purge_thds.pop_front();
   }
   n_purge_thds= 0;
@@ -1954,8 +1944,7 @@ void srv_purge_shutdown()
 		while(!srv_purge_should_exit()) {
 			ut_a(!purge_sys.paused());
 			srv_wake_purge_thread_if_not_active();
-			std::this_thread::sleep_for(
-				std::chrono::milliseconds(1));
+			purge_coordinator_task.wait();
 		}
 		purge_sys.coordinator_shutdown();
 		srv_shutdown_purge_tasks();
