@@ -1,4 +1,4 @@
-/* Copyright (c) 2008, 2011, Oracle and/or its affiliates. All rights reserved.
+/* Copyright (c) 2008, 2022, Oracle and/or its affiliates.
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License, version 2.0,
@@ -21,58 +21,67 @@
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1335 USA */
 
 /**
-  @file storage/perfschema/table_file_summary.cc
+  @file storage/perfschema/table_file_summary_by_instance.cc
   Table FILE_SUMMARY_BY_INSTANCE (implementation).
 */
 
 #include "my_global.h"
-#include "my_pthread.h"
+#include "my_thread.h"
 #include "pfs_instr_class.h"
 #include "pfs_column_types.h"
 #include "pfs_column_values.h"
 #include "table_file_summary_by_instance.h"
 #include "pfs_global.h"
+#include "pfs_buffer_container.h"
+#include "field.h"
 
 THR_LOCK table_file_summary_by_instance::m_table_lock;
 
+PFS_engine_table_share_state
+table_file_summary_by_instance::m_share_state = {
+  false /* m_checked */
+};
+
 PFS_engine_table_share
 table_file_summary_by_instance::m_share=
 {
   { C_STRING_WITH_LEN("file_summary_by_instance") },
   &pfs_truncatable_acl,
-  &table_file_summary_by_instance::create,
+  table_file_summary_by_instance::create,
   NULL, /* write_row */
   table_file_summary_by_instance::delete_all_rows,
-  NULL, /* get_row_count */
-  1000, /* records */
+  table_file_summary_by_instance::get_row_count,
   sizeof(PFS_simple_index),
   &m_table_lock,
   { C_STRING_WITH_LEN("CREATE TABLE file_summary_by_instance("
-                      "FILE_NAME VARCHAR(512) not null,"
-                      "EVENT_NAME VARCHAR(128) not null,"
-                      "OBJECT_INSTANCE_BEGIN BIGINT unsigned not null,"
-                      "COUNT_STAR BIGINT unsigned not null,"
-                      "SUM_TIMER_WAIT BIGINT unsigned not null,"
-                      "MIN_TIMER_WAIT BIGINT unsigned not null,"
-                      "AVG_TIMER_WAIT BIGINT unsigned not null,"
-                      "MAX_TIMER_WAIT BIGINT unsigned not null,"
-                      "COUNT_READ BIGINT unsigned not null,"
-                      "SUM_TIMER_READ BIGINT unsigned not null,"
-                      "MIN_TIMER_READ BIGINT unsigned not null,"
-                      "AVG_TIMER_READ BIGINT unsigned not null,"
-                      "MAX_TIMER_READ BIGINT unsigned not null,"
-                      "SUM_NUMBER_OF_BYTES_READ BIGINT not null,"
-                      "COUNT_WRITE BIGINT unsigned not null,"
-                      "SUM_TIMER_WRITE BIGINT unsigned not null,"
-                      "MIN_TIMER_WRITE BIGINT unsigned not null,"
-                      "AVG_TIMER_WRITE BIGINT unsigned not null,"
-                      "MAX_TIMER_WRITE BIGINT unsigned not null,"
-                      "SUM_NUMBER_OF_BYTES_WRITE BIGINT not null,"
-                      "COUNT_MISC BIGINT unsigned not null,"
-                      "SUM_TIMER_MISC BIGINT unsigned not null,"
-                      "MIN_TIMER_MISC BIGINT unsigned not null,"
-                      "AVG_TIMER_MISC BIGINT unsigned not null,"
-                      "MAX_TIMER_MISC BIGINT unsigned not null)") }
+                      "FILE_NAME VARCHAR(512) not null comment 'File name.',"
+                      "EVENT_NAME VARCHAR(128) not null comment 'Event name.',"
+                      "OBJECT_INSTANCE_BEGIN BIGINT unsigned not null comment 'Address in memory. Together with FILE_NAME and EVENT_NAME uniquely identifies a row.',"
+                      "COUNT_STAR BIGINT unsigned not null comment 'Number of summarized events',"
+                      "SUM_TIMER_WAIT BIGINT unsigned not null comment 'Total wait time of the summarized events that are timed.',"
+                      "MIN_TIMER_WAIT BIGINT unsigned not null comment 'Minimum wait time of the summarized events that are timed.',"
+                      "AVG_TIMER_WAIT BIGINT unsigned not null comment 'Average wait time of the summarized events that are timed.',"
+                      "MAX_TIMER_WAIT BIGINT unsigned not null comment 'Maximum wait time of the summarized events that are timed.',"
+                      "COUNT_READ BIGINT unsigned not null comment 'Number of all read operations, including FGETS, FGETC, FREAD, and READ.',"
+                      "SUM_TIMER_READ BIGINT unsigned not null comment 'Total wait time of all read operations that are timed.',"
+                      "MIN_TIMER_READ BIGINT unsigned not null comment 'Minimum wait time of all read operations that are timed.',"
+                      "AVG_TIMER_READ BIGINT unsigned not null comment 'Average wait time of all read operations that are timed.',"
+                      "MAX_TIMER_READ BIGINT unsigned not null comment 'Maximum wait time of all read operations that are timed.',"
+                      "SUM_NUMBER_OF_BYTES_READ BIGINT not null comment 'Bytes read by read operations.',"
+                      "COUNT_WRITE BIGINT unsigned not null comment 'Number of all write operations, including FPUTS, FPUTC, FPRINTF, VFPRINTF, FWRITE, and PWRITE.',"
+                      "SUM_TIMER_WRITE BIGINT unsigned not null comment 'Total wait time of all write operations that are timed.',"
+                      "MIN_TIMER_WRITE BIGINT unsigned not null comment 'Minimum wait time of all write operations that are timed.',"
+                      "AVG_TIMER_WRITE BIGINT unsigned not null comment 'Average wait time of all write operations that are timed.',"
+                      "MAX_TIMER_WRITE BIGINT unsigned not null comment 'Maximum wait time of all write operations that are timed.',"
+                      "SUM_NUMBER_OF_BYTES_WRITE BIGINT not null comment 'Bytes written by write operations.',"
+                      "COUNT_MISC BIGINT unsigned not null comment 'Number of all miscellaneous operations not counted above, including CREATE, DELETE, OPEN, CLOSE, STREAM_OPEN, STREAM_CLOSE, SEEK, TELL, FLUSH, STAT, FSTAT, CHSIZE, RENAME, and SYNC.',"
+                      "SUM_TIMER_MISC BIGINT unsigned not null comment 'Total wait time of all miscellaneous operations that are timed.',"
+                      "MIN_TIMER_MISC BIGINT unsigned not null comment 'Minimum wait time of all miscellaneous operations that are timed.',"
+                      "AVG_TIMER_MISC BIGINT unsigned not null comment 'Average wait time of all miscellaneous operations that are timed.',"
+                      "MAX_TIMER_MISC BIGINT unsigned not null comment 'Maximum wait time of all miscellaneous operations that are timed.')") },
+  false, /* m_perpetual */
+  false, /* m_optional */
+  &m_share_state
 };
 
 PFS_engine_table* table_file_summary_by_instance::create(void)
@@ -86,6 +95,12 @@ int table_file_summary_by_instance::delete_all_rows(void)
   return 0;
 }
 
+ha_rows
+table_file_summary_by_instance::get_row_count(void)
+{
+  return global_file_container.get_row_count();
+}
+
 table_file_summary_by_instance::table_file_summary_by_instance()
   : PFS_engine_table(&m_share, &m_pos),
   m_row_exists(false), m_pos(0), m_next_pos(0)
@@ -101,17 +116,14 @@ int table_file_summary_by_instance::rnd_next(void)
 {
   PFS_file *pfs;
 
-  for (m_pos.set_at(&m_next_pos);
-       m_pos.m_index < file_max;
-       m_pos.next())
+  m_pos.set_at(&m_next_pos);
+  PFS_file_iterator it= global_file_container.iterate(m_pos.m_index);
+  pfs= it.scan_next(& m_pos.m_index);
+  if (pfs != NULL)
   {
-    pfs= &file_array[m_pos.m_index];
-    if (pfs->m_lock.is_populated())
-    {
-      make_row(pfs);
-      m_next_pos.set_after(&m_pos);
-      return 0;
-    }
+    make_row(pfs);
+    m_next_pos.set_after(&m_pos);
+    return 0;
   }
 
   return HA_ERR_END_OF_FILE;
@@ -122,14 +134,15 @@ int table_file_summary_by_instance::rnd_pos(const void *pos)
   PFS_file *pfs;
 
   set_position(pos);
-  DBUG_ASSERT(m_pos.m_index < file_max);
-  pfs= &file_array[m_pos.m_index];
 
-  if (! pfs->m_lock.is_populated())
-    return HA_ERR_RECORD_DELETED;
+  pfs= global_file_container.get(m_pos.m_index);
+  if (pfs != NULL)
+  {
+    make_row(pfs);
+    return 0;
+  }
 
-  make_row(pfs);
-  return 0;
+  return HA_ERR_RECORD_DELETED;
 }
 
 /**
@@ -138,7 +151,7 @@ int table_file_summary_by_instance::rnd_pos(const void *pos)
 */
 void table_file_summary_by_instance::make_row(PFS_file *pfs)
 {
-  pfs_lock lock;
+  pfs_optimistic_state lock;
   PFS_file_class *safe_class;
 
   m_row_exists= false;
@@ -175,7 +188,7 @@ int table_file_summary_by_instance::read_row_values(TABLE *table,
     return HA_ERR_RECORD_DELETED;
 
   /* Set the null bits */
-  DBUG_ASSERT(table->s->null_bytes == 0);
+  assert(table->s->null_bytes == 0);
 
   for (; (f= *fields) ; fields++)
   {
@@ -263,7 +276,7 @@ int table_file_summary_by_instance::read_row_values(TABLE *table,
         set_field_ulonglong(f, m_row.m_io_stat.m_misc.m_waits.m_max);
         break;
       default:
-        DBUG_ASSERT(false);
+        assert(false);
       }
     }
   }
